/**
 * 
 */
package listes;

import java.util.Iterator;
import java.util.NoSuchElementException;

//import listes.Liste.Maillon;

public class Liste<E> implements IListe<E>
{

	private Maillon<E> head;
	private class ListeIterator<F> implements Iterator<E>
	{

//		private F element;
		private Maillon<F> current;
		private Maillon<F> last;
		private Maillon<F> penultimate;
		private boolean nextCalled;

		@SuppressWarnings("unchecked")
		public ListeIterator()
		{
			current = (Maillon<F>) head;
			last = null;
			penultimate = null;
			nextCalled = false;
		}

		@Override
		public boolean hasNext()
		{
			if (current == null)
			{
				return false;
			}
			return true;
		}

		/*
		 * (non-Javadoc)
		 * 
		 * @see java.util.Iterator#next()
		 */

		@SuppressWarnings("unchecked")
		@Override
		public E next() throws NoSuchElementException
		{
			if (current == null)
				throw new NoSuchElementException("No more elements");

			penultimate = last;
			last = current;
			current = current.getNext();
			nextCalled = true; // Used in remove
			return (E) last.getData();
		}

		@SuppressWarnings("unchecked")
		public void remove() throws IllegalStateException
		{
			if ((last == null) || (! nextCalled) ) 
				throw new IllegalStateException();
			
			if (last == head)
			{
				head = (Maillon<E>) current;
			}
			else
			{
				penultimate.next = current;
			}
			nextCalled = false; // Avoids calling remove twice ... 
		}
	}

	
	public static class Maillon<E>
	{
		private E data;
		private Maillon<E> next;

		public Maillon(E data, Maillon<E> next)
		{
			this.data = data;
			this.next = next;
		}

		public Maillon(E data)
		{
			this.data = data;
			this.next = null;
		}

		public E getData()
		{
			return data;
		}

		public Maillon<E> getNext()
		{
			return next;
		}

		public void setNext(Maillon<E> next)
		{
			this.next = next;
		}
	}

	public Liste()
	{
		this.head = null;
	}

	public Liste(Liste<E> l)
	{
		this.head = null;
		Iterator<E> iter = l.iterator();
		while (iter.hasNext())
		{
			add(iter.next());
		}
	}

	public void add(E elt) throws NullPointerException
	{
		if (elt == null)
			throw new NullPointerException();

		Maillon<E> nouv = new Maillon<E>(elt); 
		if (head == null)
		{
			head = nouv;
		}
		else
		{
			Maillon<E> cur = head;
			while (cur.next != null)
				cur = cur.next;
			cur.next = nouv;
		}
	}

	public void insert(E elt) throws NullPointerException
	{
		if (elt == null)
//			return false;
			throw new NullPointerException();
		insert(elt, 0);
	}

	public boolean insert(E elt, int index)
	{
		if (elt == null)
			return false;
//			throw new NullPointerException();

		if (index < 0)
			return false;

		Maillon<E> m = new Maillon<E>(elt);

		// Special case if 0 
		if (index == 0)
		{
			m.next = head;
			head = m;
			return true;
		}

		// Empty list
		if (head == null)
			return false;
		
		Maillon<E> cur = head;

		// Find insert position
		while (--index != 0)
		{
			cur = cur.next;
			if (cur == null)
				return false; 
		}
		// link it
		m.next = cur.next;
		cur.next = m;
		return true;
	}

  	public Iterator<E> iterator()
	{
		// Use the ad hoc class ...
		return new ListeIterator<E>();
	}

	@Override
	public boolean remove(E elt)
	{
		Iterator<E> iter = iterator();
		while (iter.hasNext())
		{
			E el = iter.next();
			if (el.equals(elt))
			{
				iter.remove();
				return true;
			}
		}
		return false;
	}

	@Override
	public boolean removeAll(E elt)
	{
		boolean resu = false;
		Iterator<E> iter = iterator();
		while (iter.hasNext())
		{
			E el = iter.next();
			if (el.equals(elt))
			{
				resu = true;
				iter.remove();
			}
		}
		return resu;
	}

	@Override
	public int size()
	{
		int compteur = 0;
		for (@SuppressWarnings("unused") E el : this)
		{
			compteur++;
		}
		return compteur;
	}

	@Override
	public void clear()
	{
		head = null;
	}

	@Override
	public boolean empty()
	{
		return head == null;
	}

	@Override
	public int hashCode()
	{
		int resu = 1; 
		Iterator<E> iter = iterator();
		while (iter.hasNext())
		{
			E el = iter.next();
			// Use same hash method for hashcodes consistency
			resu = 31 * resu + el.hashCode();
		}
		return resu;
	}

	@Override
	public boolean equals(Object obj)
	{
		if (obj instanceof Iterable<?>)
		{
			Iterable<?> otherlist = (Iterable<?>) obj;
			
			Iterator<E> iter = iterator(); // Local iterator 
			Iterator<?> otheriter = otherlist.iterator(); // Local iterator 
			
			while (iter.hasNext())
			{
				E myelem = iter.next();
				if (! otheriter.hasNext()) // Other list have less elements
					return false;
				if (! myelem.equals(otheriter.next())) // Elements don't match ...
						return false;
			}
			if (otheriter.hasNext()) // Other list have more elements
				return false;
			
			// all good
			return true;
		}
		return false;
	}

	@Override
	public String toString()
	{
		// Apparently we want -> as separator and [ ] around like this : [elem1->elem2->etc]
		
		if (head == null)
			return "[]";
		
		StringBuilder sb = new StringBuilder();
		Iterator<E> iter = iterator();
		while (iter.hasNext())
		{
			sb.append("->");
			sb.append(iter.next());
		}
		// Skip first ->
		return "[" + sb.substring(2) + "]";
	}


}
